/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * Ca.java,v 1.1.1.1 2003/09/11 14:42:49 boucher Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */

package gov.aps.epics.jca;

import gov.aps.epics.jca.dbr.*;
import gov.aps.epics.jca.jni.JNICa;

import java.io.*;

/**
 * 
 *
 *
 *
 * @author <a href="mailto:boucher@aps.anl.gov">Eric Boucher</a> 
 */
public class Ca implements CaConstants {

  // CA

  /**
   * Returns the CA version.
   */
  static public int    getVersion()       { return JNICa.getVersion(); }
  
  /**
   * Returns the CA revision.
   */
  static public int    getRevision()      { return JNICa.getRevision(); }
  
  /**
   * Returns the CA modification.
   */
  static public int    getModification()  { return JNICa.getModification(); }
  
  /**
   * Returns the CA version's String.
   */
  static public String getVersionString() { return JNICa.getVersionString(); }
  
  /**
   * Returns the CA release String.
   */
  static public String getReleaseString() { return JNICa.getReleaseString(); }


  // CONTEXT

  /**
   * Creates a new Context with a default name, preemptive callbacks disabled and an internal thread control mode.
   *
   */
  static public Context createContext() throws CaException {
    return createContext(null, DISABLE_PREEMPTIVE_CALLBACK, THREAD_CTRL_INTERNAL);
  }

  /**
   * Creates a new Context with a specified name, preemptive callbacks disabled and an internal thread control mode.
   *
   * @param name the Context name.
   *
   */
  static public Context createContext(String name) throws CaException {
    return createContext(name, DISABLE_PREEMPTIVE_CALLBACK, THREAD_CTRL_INTERNAL);
  }

  /**
   * Creates a new Context with a default name, a specified callback mode and an internal thread control mode.
   *
   * @param callback the callback mode: {@link gov.aps.epics.jca.CaConstants#DISABLE_PREEMPTIVE_CALLBACK DISABLE_PREEMPTIVE_CALLBACK} or {@link gov.aps.epics.jca.CaConstants#ENABLE_PREEMPTIVE_CALLBACK ENABLE_PREEMPTIVE_CALLBACK}.
   *
   */
  static public Context createContext(int callback) throws CaException {
    return createContext(null, callback, THREAD_CTRL_INTERNAL);
  }

  /**
   * Creates a new Context with a specified name, a specified callback mode and an internal thread control mode.
   *
   * @param name the Context name.
   * @param callback the callback mode: {@link gov.aps.epics.jca.CaConstants#DISABLE_PREEMPTIVE_CALLBACK DISABLE_PREEMPTIVE_CALLBACK} or {@link gov.aps.epics.jca.CaConstants#ENABLE_PREEMPTIVE_CALLBACK ENABLE_PREEMPTIVE_CALLBACK}.
   *
   */
  static public Context createContext(String name, int callback) throws CaException {
    return createContext(name, callback, THREAD_CTRL_INTERNAL);
  }

  /**
   * Creates a new Context with a default name, a specified callback mode and a specified thread control mode.
   *
   * @param name the Context name.
   * @param thread the thread control mode: {@link gov.aps.epics.jca.CaConstants#THREAD_CTRL_EXTERNAL THREAD_CTRL_EXTERNAL} or {@link gov.aps.epics.jca.CaConstants#THREAD_CTRL_INTERNAL THREAD_CTRL_INTERNAL}.
   *
   */
  static public Context createContext(int callback, int thread) throws CaException {
    return createContext(null, callback, thread);
  }

  /**
   * Creates a new Context with a specified name, a specified callback mode and a specified thread control mode.
   *
   * @param name the Context name.
   * @param callback the callback mode: {@link gov.aps.epics.jca.CaConstants#DISABLE_PREEMPTIVE_CALLBACK DISABLE_PREEMPTIVE_CALLBACK} or {@link gov.aps.epics.jca.CaConstants#ENABLE_PREEMPTIVE_CALLBACK ENABLE_PREEMPTIVE_CALLBACK}.
   * @param thread the thread control mode: {@link gov.aps.epics.jca.CaConstants#THREAD_CTRL_EXTERNAL THREAD_CTRL_EXTERNAL} or {@link gov.aps.epics.jca.CaConstants#THREAD_CTRL_INTERNAL THREAD_CTRL_INTERNAL}.
   *
   * @return the new Context.
   */
  static public Context createContext(String name, int callback, int thread) throws CaException {
    return JNICa.createContext(name,callback,thread);
  }

  /**
   * Returns the list of all contexts created .
   */
  static public Context[] getContextList() {
    return Context.getContextList();
  }


  /**
   * Returns the String representing the Context's callback mode.
   *
   * @param context the context.
   * @return the representing String.
   * @see Context
   */
  static public String getCallbackModeString(Context context) {
    return getCallbackModeString(context.getCallbackMode());
  }

  /**
   * Returns the callback mode corresponding to the mode string.
   *
   * @param callback the callback mode name.
   * @return the callback mode or -1 if the name does not correspond to a valid mode.
   * @see Context
   */
  static public int getCallbackMode(String callback) {
    for(int t=CALLBACK_MODE_FIRST; t<=CALLBACK_MODE_LAST; ++t) {
      if(callback.equals(CALLBACK_MODE_STRING[t])) return t;
    }
    return -1;
  }

  /**
   * Returns the String representing the Context's callback mode.
   *
   * @param callback the callback mode.
   * @return the representing String.
   * @see Context
   */
  static public String getCallbackModeString(int callback) {
    if(callback<CALLBACK_MODE_FIRST || callback>CALLBACK_MODE_LAST) return "UNKNOWN_MODE";
    return CALLBACK_MODE_STRING[callback];
  }

  
  /**
   * Returns the String representing the Context's thread control mode.
   *
   * @param context the context.
   * @return the representing String.
   * @see Context
   */
  static public String getThreadCtrlModeString(Context context) {
    return getThreadCtrlModeString(context.getThreadCtrlMode());
  }
  
  /**
   * Returns the thread mode corresponding to the mode string.
   *
   * @param thread the thread mode name.
   * @return the thread mode or -1 if the name does not correspond to a valid mode.
   * @see Context
   */
  static public int getThreadCtrlMode(String thread) {
    for(int t=THREAD_CTRL_FIRST; t<=THREAD_CTRL_LAST; ++t) {
      if(thread.equals(THREAD_CTRL_STRING[t])) return t;
    }
    return -1;
  }
  
  /**
   * Returns the String representing the Context's thread control mode.
   *
   * @param thread the thread control mode.
   * @return the representing String.
   * @see Context
   */
  static public String getThreadCtrlModeString(int thread) {
    if(thread<THREAD_CTRL_FIRST || thread>THREAD_CTRL_LAST) return "UNKNOWN_MODE";
    return THREAD_CTRL_STRING[thread];
  }


  // INFO

  /**
   * Prints detailed information about the provided Context to the standard output stream.
   *
   * @param ctx the context.
   */
  static public void printInfo(Context ctx) {
    ctx.printInfo(System.out);
  }

  /**
   * Prints detailed information about the provided Context to the specified output stream.
   *
   * @param ctx the context.
   * @param out the output stream.
   */
  static public void printInfo(Context ctx, PrintStream out) {
    ctx.printInfo(out);
  }

  /**
   * Prints detailed information about the provided DBR object to the standard output stream.
   *
   * @param dbr the context.
   */
  static public void printInfo(DBR dbr) {
    dbr.printInfo(System.out);
  }
  
  /**
   * Prints detailed information about the provided DBR object to the specified output stream.
   *
   * @param dbr the context.
   * @param out the output stream.
   */
  static public void printInfo(DBR dbr, PrintStream out) {
    dbr.printInfo(out);
  }

  /**
   * Prints detailed information about the provided Channel to the standard output stream.
   *
   * @param dbr the context.
   */
  static public void printInfo(Channel ch) {
    ch.printInfo(System.out);
  }
  
  /**
   * Prints detailed information about the provided Channel to the specified output stream.
   *
   * @param dbr the context.
   * @param out the output stream.
   */
  static public void printInfo(Channel ch, PrintStream out) {
    ch.printInfo(out);
  }


  // CA_OP

  /**
   * Returns the String representing the CA_OP code.
   *
   * @param ca_op the CA_OP_XXX code.
   * @return the representing String.
   * @see ContextExceptionEvent
   */
  static public String getOPString(int ca_op) {
    if(ca_op<CA_OP_FIRST || ca_op>CA_OP_LAST) return "CA_OP_UNKNOWN";
    return CA_OP_STRING[ca_op];
  }


  // CS_XXX
  /**
   * Returns the String representing the CS_XXX code.
   *
   * @param cs_xxx the CS_XXX code.
   * @return the representing String.
   * @see Channel#getState
   */
  static public String getConnectionStateString(int cs_xxx) {
    if(cs_xxx<CS_FIRST || cs_xxx>CS_LAST) return "CS_UNKNOWN";
    return CS_STRING[cs_xxx];
  }

  // ECA_MESSAGE

  /**
   * Returns the severity value of the ECA_XXX code.
   * 
   * @param eca_message the ECA_XXX code.
   * @return the severity valule. One of CA_K_XXX.
   * @see CaConstants
   */
  static public int getECACodeSeverity(int eca_message) {
     return  ( ( (eca_message) & CA_M_SEVERITY )    >> CA_V_SEVERITY );
  }

  /**
   * Returns the message number of the ECA_XXX code.
   *
   * @param eca_message the ECA_XXX code.
   * @return the message number.
   * @see CaConstants
   */
  static public int getECACodeNo(int eca_message) {
    return ( ( (eca_message) & CA_M_MSG_NO )  >> CA_V_MSG_NO );
  }



  /**
   * Returns the String representing the ECA_XXX code.
   * 
   * @param eca_message the ECA_XXX code.
   * @return the representing String.
   * @see CaConstants
   */
  static public String getECACodeMessage(int eca_message) {
    int no= getECACodeNo(eca_message);
    if(no<ECA_FIRST || no>ECA_LAST) return "Unknown message";
    return ECA_MESSAGE_STRING[no];
  }

  // DBE

  /**
   * Determines wether the monitor mask includes the DBE_VALUE flag.
   *
   * @param mask the mask value.
   * @return true if the mask includes DBE_VALUE, false otherwise.
   * @see CaConstants
   */
  static public boolean isDBE_VALUE(int mask) {
    return (mask & DBE_VALUE)!=0;
  }
  
  /**
   * Determines wether the monitor mask includes the DBE_LOG flag.
   *
   * @param mask the mask value.
   * @return true if the mask includes DBE_LOG, false otherwise.
   * @see CaConstants
   */
  static public boolean isDBE_LOG(int mask) {
    return (mask & DBE_LOG)!=0;
  }

  /**
   * Determines wether the monitor mask includes the DBE_ALARM flag.
   *
   * @param mask the mask value.
   * @return true if the mask includes DBE_LOG, false otherwise.
   * @see CaConstants
   */
  static public boolean isDBE_ALARM(int mask) {
    return (mask & DBE_ALARM)!=0;
  }

  // DBR

  /**
   * Returns the String representing the type of the DBR object.
   *
   * @param dbr the DBR object.
   * @return the type's String.
   * @see CaConstants
   */
  static public String getDBRTypeString(DBR dbr) {
    return getDBRTypeString(dbr.getType());
  }

  /**
   * Returns the String representing the dbr type value.
   *
   * @param dbr the DBR type value. One of DBR_XXX.
   * @return the type's String.
   * @see CaConstants
   */
  static public String getDBRTypeString(int type) {
    if(type<DBR_FIRST || type>DBR_LAST) return "UNKNOWN_TYPE";
    return DBR_TYPE_STRING[type];
  }


  // VALUE

  /**
   * Helper method to print an array of byte as a comma separated list to the standard output stream.
   *
   * @param value the array to print.
   */
  static public void printValue(byte[] value) {
    printValue(value, System.out);
  }

  /**
   * Helper method to print an array of byte as a comma separated list to the specified output stream.
   *
   * @param value the array to print.
   * @param out the output stream.
   */
  static public void printValue(byte[] value, PrintStream out) {
    for(int t=0; t<value.length; ++t) {
      if(t>0) out.print(",");
      out.print(value[t]);
    }
  }
  
  /**
   * Helper method to print an array of short as a comma separated list to the standard output stream.
   *
   * @param value the array to print.
   */
  static public void printValue(short[] value) {
    printValue(value, System.out);
  }

  /**
   * Helper method to print an array of short as a comma separated list to the specified output stream.
   *
   * @param value the array to print.
   * @param out the output stream.
   */
  static public void printValue(short[] value, PrintStream out) {
    for(int t=0; t<value.length; ++t) {
      if(t>0) out.print(",");
      out.print(value[t]);
    }
  }
  
  /**
   * Helper method to print an array of int as a comma separated list to the standard output stream.
   *
   * @param value the array to print.
   */
  static public void printValue(int[] value) {
    printValue(value, System.out);
  }

  /**
   * Helper method to print an array of int as a comma separated list to the specified output stream.
   *
   * @param value the array to print.
   * @param out the output stream.
   */
  static public void printValue(int[] value, PrintStream out) {
    for(int t=0; t<value.length; ++t) {
      if(t>0) out.print(",");
      out.print(value[t]);
    }
  }
  
  /**
   * Helper method to print an array of long as a comma separated list to the standard output stream.
   *
   * @param value the array to print.
   */
  static public void printValue(long[] value) {
    printValue(value, System.out);
  }

  /**
   * Helper method to print an array of long as a comma separated list to the specified output stream.
   *
   * @param value the array to print.
   * @param out the output stream.
   */
  static public void printValue(long[] value, PrintStream out) {
    for(int t=0; t<value.length; ++t) {
      if(t>0) out.print(",");
      out.print(value[t]);
    }
  }
  
  /**
   * Helper method to print an array of float as a comma separated list to the standard output stream.
   *
   * @param value the array to print.
   */
  static public void printValue(float[] value) {
    printValue(value, System.out);
  }

  /**
   * Helper method to print an array of float as a comma separated list to the specified output stream.
   *
   * @param value the array to print.
   * @param out the output stream.
   */
  static public void printValue(float[] value, PrintStream out) {
    for(int t=0; t<value.length; ++t) {
      if(t>0) out.print(",");
      out.print(value[t]);
    }
  }
  
  /**
   * Helper method to print an array of double as a comma separated list to the standard output stream.
   *
   * @param value the array to print.
   */
  static public void printValue(double[] value) {
    printValue(value, System.out);
  }

  /**
   * Helper method to print an array of double as a comma separated list to the specified output stream.
   *
   * @param value the array to print.
   * @param out the output stream.
   */
  static public void printValue(double[] value, PrintStream out) {
    for(int t=0; t<value.length; ++t) {
      if(t>0) out.print(",");
      out.print(value[t]);
    }
  }
  
  /**
   * Helper method to print an array of String as a comma separated list to the standard output stream.
   *
   * @param value the array to print.
   */
  static public void printValue(String[] value) {
    printValue(value, System.out);
  }

  /**
   * Helper method to print an array of String as a comma separated list to the specified output stream.
   *
   * @param value the array to print.
   * @param out the output stream.
   */
  static public void printValue(String[] value, PrintStream out) {
    for(int t=0; t<value.length; ++t) {
      if(t>0) out.print(",");
      out.print(value[t]);
    }
  }


}
