/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * JNI.cpp,v 1.4 2005/12/05 17:24:52 jba Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */


#include "gov_aps_epics_jca_jni_JNI.h"

#include <epicsVersion.h>
#include <stdlib.h>
#include <string.h>
#include <cadef.h>
#include "gov_aps_epics_jca_CaConstants.h"



static JavaVM* _jvm=NULL;


static void checkstatus(JNIEnv* env, int code) {
  if (code==ECA_NORMAL) return;
  jclass clazz= env->FindClass("gov/aps/epics/jca/CaException");
  jmethodID mid= env->GetMethodID(clazz, "<init>", "(I)V");

  jthrowable cae= (jthrowable) env->NewObject(clazz,mid,code);
  env->Throw(cae);
}

JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ca_1getVersion(JNIEnv* env, jclass) {
  return EPICS_VERSION;
}

JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ca_1getRevision(JNIEnv* env, jclass) {
  return EPICS_REVISION;
}

JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ca_1getModification(JNIEnv* env, jclass) {
  return EPICS_MODIFICATION;
}

JNIEXPORT jstring JNICALL Java_gov_aps_epics_jca_jni_JNI__1ca_1getVersionString(JNIEnv *env, jclass) {
  return env->NewStringUTF(EPICS_VERSION_STRING);
}

JNIEXPORT jstring JNICALL Java_gov_aps_epics_jca_jni_JNI__1ca_1getReleaseString(JNIEnv *env, jclass) {
  return env->NewStringUTF(epicsReleaseVersion);
}



#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
static epicsThreadPrivateId messageCallbackID;
static epicsThreadPrivateId exceptionCallbackID;

#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
static jobject messageCallback;
static jobject exceptionCallback;

#endif



extern "C" {
  static int theMessageCallback(const char* pFormat, va_list args) {

#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    jobject callback= (jobject) epicsThreadPrivateGet(messageCallbackID);

#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
    jobject callback= messageCallback;

#endif

    if (callback==NULL) return ECA_NORMAL;

    JNIEnv* env;
    _jvm->AttachCurrentThread((void**)&env, NULL);

    char buf[1024];
    sprintf(buf, pFormat, args);
    jstring msg= env->NewStringUTF(buf);

    jclass clazz= env->FindClass("gov/aps/epics/jca/jni/JNIContextMessageCallback");
    jmethodID methodID= env->GetMethodID(clazz, "fire", "(Ljava/lang/String;)V");


    env->CallVoidMethod(callback, methodID,  msg);

#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    _jvm->DetachCurrentThread();
#endif

    return ECA_NORMAL;
  }


  static void theExceptionCallback(struct exception_handler_args args) {

#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    jobject callback= (jobject) epicsThreadPrivateGet(exceptionCallbackID);

#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
    jobject callback= exceptionCallback;

#endif

    if (callback==NULL) return;

    JNIEnv* env;
    _jvm->AttachCurrentThread((void**)&env, NULL);

    jclass clazz= env->FindClass("gov/aps/epics/jca/jni/JNIContextExceptionCallback");
    jmethodID methodID= env->GetMethodID(clazz, "fire", "(IIIIIILjava/lang/String;Ljava/lang/String;I)V");

    jstring ctxtInfo=NULL;
    if (args.ctx!=NULL) ctxtInfo= env->NewStringUTF(args.ctx);
    jstring file=NULL;
    if (args.pFile!=NULL) file= env->NewStringUTF(args.pFile);


    env->CallVoidMethod(callback, methodID,  (jint) args.chid,
                        (jint) args.type,
                        (jint) args.count,
                        (jint) args.addr,
                        (jint) args.stat,
                        (jint) args.op,
                        (jstring) ctxtInfo,
                        (jstring) file,
                        (jint) args.lineNo );

#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    _jvm->DetachCurrentThread();
#endif

  }
} /* extern "C" */





JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1contextCreate(JNIEnv *env, jclass, jint mode) {


#if (EPICS_VERSION==3) && (EPICS_REVISION==14)

  if (_jvm==NULL) {
    env->GetJavaVM(&_jvm);
    messageCallbackID= epicsThreadPrivateCreate();
    exceptionCallbackID= epicsThreadPrivateCreate();
  }

  ca_preemptive_callback_select ca_mode;

  switch (mode) {
  case gov_aps_epics_jca_CaConstants_ENABLE_PREEMPTIVE_CALLBACK: ca_mode= ca_enable_preemptive_callback; break;
  default: ca_mode= ca_disable_preemptive_callback;
  }

  checkstatus( env, ca_context_create(ca_mode) );

  return(jint) ca_current_context();

#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)

  if (_jvm!=0 || mode!=jca_CaConstants_DISABLE_PREEMPTIVE_CALLBACK) {
    return 0;
  }

  env->GetJavaVM(&_jvm);
  checkstatus( env,  ca_task_initialize() );

  return 0;
#endif

}


JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1contextDestroy(JNIEnv *env, jclass) {
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)

  jobject msg= (jobject) epicsThreadPrivateGet(messageCallbackID);
  jobject excpt= (jobject) epicsThreadPrivateGet(exceptionCallbackID);

  ca_context_destroy() ;

  if (msg!=NULL) env->DeleteGlobalRef(msg);
  if (excpt!=NULL) env->DeleteGlobalRef(excpt);


#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)

  if (messageCallback!=NULL) env->DeleteGlobalRef(messageCallback);
  if (exceptionCallback!=NULL) env->DeleteGlobalRef(exceptionCallback);

  env, ca_task_exit() ;
  _jvm=NULL;

#endif

}


JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1setMessageCallback(JNIEnv *env, jclass, jobject l) {

#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
  jobject old= (jobject) epicsThreadPrivateGet(messageCallbackID);
#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
  jobject old= messageCallback;
#endif

  if (l!=NULL) l= env->NewGlobalRef(l);

#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
  epicsThreadPrivateSet(messageCallbackID, (void*) l);
#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
  messageCallback=l;
#endif

  if (old!=NULL) env->DeleteGlobalRef(old);

  if (l!=NULL) {
    checkstatus( env, ca_replace_printf_handler(&theMessageCallback) );
  } else {
    checkstatus( env, ca_replace_printf_handler(NULL) );
  }

}


JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1setExceptionCallback(JNIEnv *env, jclass, jobject l) {
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
  jobject old= (jobject) epicsThreadPrivateGet(exceptionCallbackID);
#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
  jobject old= exceptionCallback;
#endif

  if (l!=NULL) l= env->NewGlobalRef(l);
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
  epicsThreadPrivateSet(exceptionCallbackID, (void*) l);
#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
  exceptionCallback=l;
#endif

  if (old!=NULL) env->DeleteGlobalRef(old);

  if (l!=NULL) {
    checkstatus( env, ca_add_exception_event(&theExceptionCallback, NULL) );
  } else {
    checkstatus( env, ca_add_exception_event(NULL, NULL) );
  }

}





JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1pendIO(JNIEnv *env, jclass, jdouble timeout) {
  checkstatus( env, ca_pend_io(timeout) );
}

JNIEXPORT jboolean JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1testIO(JNIEnv *env, jclass) {
  int status= ca_test_io();
  checkstatus( env, status );
  return status==ECA_IOINPROGRESS;
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1pendEvent(JNIEnv* env, jclass, jdouble timeout) {
  checkstatus( env, ca_pend_event(timeout) );
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1poll(JNIEnv* env, jclass) {
  checkstatus( env, ca_poll() );
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1flushIO(JNIEnv* env, jclass) {
  checkstatus( env, ca_flush_io() );
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ctxt_1attachThread(JNIEnv *env, jclass, jint ctxtID) {
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
  checkstatus( env, ca_attach_context( (ca_client_context*)ctxtID ) );
#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
  return;
#endif

}


/**************************************************
 *
 *          CHANNEL
 *
 **************************************************/

struct usrarg {
  jobject cnxCallback;
  jobject accessCallback;
  usrarg() : cnxCallback(NULL), accessCallback(NULL) {
  }
};


#define CHID(pchid) (*((chid*)pchid))
#define PCHID(pchid) ((chid*)pchid)

#define CNXCALLBACK(chid) ((jobject)((usrarg*)ca_puser(chid))->cnxCallback)
#define SET_CNXCALLBACK(chid, cb) ((usrarg*)ca_puser(chid))->cnxCallback=cb
#define ACCESSCALLBACK(chid) ((jobject)((usrarg*)ca_puser(chid))->accessCallback)
#define SET_ACCESSCALLBACK(chid, cb) ((usrarg*)ca_puser(chid))->accessCallback=cb




extern "C" {
  static void theConnectionCallback(struct connection_handler_args args) {
    if (CNXCALLBACK(args.chid)==NULL) return;

    JNIEnv* env;
    _jvm->AttachCurrentThread((void**)&env, NULL);


    jclass clazz= env->FindClass("gov/aps/epics/jca/jni/JNIConnectionCallback");
    jmethodID methodID= env->GetMethodID(clazz, "fire", "(Z)V");

    env->CallVoidMethod(CNXCALLBACK(args.chid), methodID, args.op==CA_OP_CONN_UP);

#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    _jvm->DetachCurrentThread();
#endif
}

  static void theAccessRightsCallback(struct access_rights_handler_args args) {
    if (ACCESSCALLBACK(args.chid)==NULL) return;

    JNIEnv* env;
    _jvm->AttachCurrentThread((void**)&env, NULL);


    jclass clazz= env->FindClass("gov/aps/epics/jca/jni/JNIAccessRightsCallback");
    jmethodID methodID= env->GetMethodID(clazz, "fire", "(ZZ)V");

    env->CallVoidMethod(ACCESSCALLBACK(args.chid), methodID, args.ar.read_access==1, args.ar.write_access==1);
  
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    _jvm->DetachCurrentThread();
#endif

}

  static void theArrayPutCallback(struct event_handler_args args) {
    JNIEnv* env;
    _jvm->AttachCurrentThread((void**)&env, NULL);


    jclass clazz= env->FindClass("gov/aps/epics/jca/jni/JNIPutCallback");
    jmethodID methodID= env->GetMethodID(clazz, "fire", "(IIII)V");

    env->CallVoidMethod((jobject) args.usr, methodID, args.type, args.count, (jint) args.dbr, args.status);
    env->DeleteGlobalRef((jobject)args.usr);
    
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    _jvm->DetachCurrentThread();
#endif

  }

  static void theArrayGetCallback(struct event_handler_args args) {
    JNIEnv* env;
    _jvm->AttachCurrentThread((void**)&env, NULL);


    jclass clazz= env->FindClass("gov/aps/epics/jca/jni/JNIGetCallback");
    jmethodID methodID= env->GetMethodID(clazz, "fire", "(IIII)V");

    env->CallVoidMethod((jobject) args.usr, methodID, args.type, args.count, (jint) args.dbr, args.status);
    env->DeleteGlobalRef((jobject)args.usr);
  
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    _jvm->DetachCurrentThread();
#endif

  }


  static void theArrayMonitorCallback(struct event_handler_args args) {
    JNIEnv* env;
    _jvm->AttachCurrentThread((void**)&env, NULL);


    jclass clazz= env->FindClass("gov/aps/epics/jca/jni/JNIMonitorCallback");
    jmethodID methodID= env->GetMethodID(clazz, "fire", "(IIII)V");

    env->CallVoidMethod((jobject) args.usr, methodID, args.type, args.count, (jint) args.dbr, args.status);
  
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    _jvm->DetachCurrentThread();
#endif

  }
} /* extern "C" */



JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1channelCreate(JNIEnv *env, jclass, jstring name, jobject l) {
  int status;
  chid* pchid= new chid;
  usrarg* pusrarg= new usrarg;

  const char* pvname= env->GetStringUTFChars(name,NULL);

  if (l!=NULL) {
    pusrarg->cnxCallback= env->NewGlobalRef(l);
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    status= ca_create_channel(pvname, &theConnectionCallback, (void*) pusrarg, 0, PCHID(pchid));
#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
    status= ca_search_and_connect(pvname,  PCHID(pchid), &theConnectionCallback, (void*) pusrarg);
#endif
  } else {
#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
    status= ca_create_channel(pvname, NULL, (void*) pusrarg, 0, PCHID(pchid));
#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
    status= ca_search_and_connect(pvname,  PCHID(pchid), NULL, (void*) pusrarg);
#endif
  }

  env->ReleaseStringUTFChars(name, pvname);

  checkstatus( env, status );

  if (status==ECA_NORMAL) return(jint)pchid;

  delete pchid;
  delete pusrarg;
  return 0;
}


JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1setConnectionCallback(JNIEnv *env, jclass, jint channelID, jobject l) {
  jobject oldcb= CNXCALLBACK(CHID(channelID));

  int status;
  if (l!=NULL) {
    SET_CNXCALLBACK(CHID(channelID), env->NewGlobalRef(l));
    status= ca_change_connection_event(CHID(channelID), &theConnectionCallback);
  } else {
    SET_CNXCALLBACK(CHID(channelID), NULL);
    status= ca_change_connection_event(CHID(channelID), NULL);
  }

  if (oldcb!=NULL) env->DeleteGlobalRef(oldcb);

  checkstatus( env, status );
}


JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1setAccessRightsCallback(JNIEnv *env, jclass, jint channelID, jobject l) {
  if (ACCESSCALLBACK(CHID(channelID))!=NULL) {
    env->DeleteGlobalRef(ACCESSCALLBACK(CHID(channelID)));
  }

  if (l!=NULL) {
    SET_ACCESSCALLBACK(CHID(channelID), env->NewGlobalRef(l));
    checkstatus( env, ca_replace_access_rights_event(CHID(channelID), &theAccessRightsCallback) );
  } else {
    SET_ACCESSCALLBACK(CHID(channelID), NULL);
    checkstatus( env, ca_replace_access_rights_event(CHID(channelID), NULL) );
  }
}


JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1channelDestroy(JNIEnv *env, jclass, jint channelID) {

  if (CNXCALLBACK(CHID(channelID))!=NULL) env->DeleteGlobalRef(CNXCALLBACK(CHID(channelID)));

  delete (usrarg*)ca_puser(CHID(channelID));

  checkstatus( env, ca_clear_channel(CHID(channelID)) );

  delete PCHID(channelID);
}


JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1getFieldType(JNIEnv* env, jclass, jint channelID) {
  return(jint) ca_field_type(CHID(channelID));
}

JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1getElementCount(JNIEnv* env, jclass, jint channelID) {
  return(jint) ca_element_count(CHID(channelID));
}

JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1getState(JNIEnv* env, jclass, jint channelID) {
  return(jint) ca_state(CHID(channelID));
}

JNIEXPORT jstring JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1getHostName(JNIEnv *env, jclass, jint channelID) {
  return env->NewStringUTF(ca_host_name(CHID(channelID)));
}

JNIEXPORT jboolean JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1getReadAccess(JNIEnv* env, jclass, jint channelID) {
  return ca_read_access(CHID(channelID));
}

JNIEXPORT jboolean JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1getWriteAccess(JNIEnv* env, jclass, jint channelID) {
  return ca_write_access(CHID(channelID));
}



JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1arrayPut(JNIEnv *env, jclass, jint type, jint length, jint channelID, jint dbrID) {
  checkstatus( env,  ca_array_put(type,length, CHID(channelID), (void*)dbrID) );
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1arrayPutCallback(JNIEnv *env, jclass, jint type, jint length, jint channelID, jint dbrID, jobject callback) {
  jobject cb= env->NewGlobalRef(callback);
  int status= ca_array_put_callback(type,length,CHID(channelID), (void*)dbrID, &theArrayPutCallback, cb);
  checkstatus( env, status  );

  if (status!=ECA_NORMAL) env->DeleteGlobalRef(cb);
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1arrayGet(JNIEnv *env, jclass, jint type, jint length, jint channelID, jint dbrID) {
  checkstatus( env, ca_array_get(type,length, CHID(channelID), (void*)dbrID) );
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1arrayGetCallback(JNIEnv *env, jclass, jint type, jint length, jint channelID, jobject callback) {
  jobject cb= env->NewGlobalRef(callback);
  checkstatus( env,  ca_array_get_callback(type,length, CHID(channelID), &theArrayGetCallback, cb) );
}


struct MonitorID {
  jobject callback;
  evid    pevid;
};


JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1addMonitor(JNIEnv *env, jclass, jint type, jint length, jint channelID, jobject callback, jint mask) {
  MonitorID* monitorID= new MonitorID;
  monitorID->callback= env->NewGlobalRef(callback);
  checkstatus( env, ca_add_masked_array_event(type,length,CHID(channelID), &theArrayMonitorCallback, monitorID->callback,0,0,0,&monitorID->pevid,mask));
  return(jint)monitorID;
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1ch_1clearMonitor(JNIEnv *env, jclass, jint monitorID) {
  checkstatus( env, ca_clear_event(((MonitorID*)monitorID)->pevid) );
  env->DeleteGlobalRef(((MonitorID*)monitorID)->callback);
}




/********************************************
 *
 *                 DBR
 *
 ********************************************/

jobject newByte(JNIEnv* env, jbyte value) {
  jclass clazz= env->FindClass("java/lang/Byte");
  jmethodID init= env->GetMethodID(clazz, "<init>", "(B)V");
  return env->NewObject(clazz, init, value);
}

jobject newShort(JNIEnv* env, jshort value) {
  jclass clazz= env->FindClass("java/lang/Short");
  jmethodID init= env->GetMethodID(clazz, "<init>", "(S)V");
  return env->NewObject(clazz, init, value);
}

jobject newInt(JNIEnv* env, jint value) {
  jclass clazz= env->FindClass("java/lang/Integer");
  jmethodID init= env->GetMethodID(clazz, "<init>", "(I)V");
  return env->NewObject(clazz, init, value);
}

jobject newFloat(JNIEnv* env, jfloat value) {
  jclass clazz= env->FindClass("java/lang/Float");
  jmethodID init= env->GetMethodID(clazz, "<init>", "(F)V");
  return env->NewObject(clazz, init, value);
}

jobject newDouble(JNIEnv* env, jdouble value) {
  jclass clazz= env->FindClass("java/lang/Double");
  jmethodID init= env->GetMethodID(clazz, "<init>", "(D)V");
  return env->NewObject(clazz, init, value);
}


#if (EPICS_VERSION==3) && (EPICS_REVISION==14)
jobject newTimeStamp(JNIEnv* env, epicsTimeStamp stamp) {
#elif (EPICS_VERSION==3) && (EPICS_REVISION<14)
jobject newTimeStamp(JNIEnv* env, TS_STAMP stamp) {
#endif
    jclass clazz= env->FindClass("gov/aps/epics/jca/dbr/TimeStamp");
  jmethodID  init= env->GetMethodID(clazz, "<init>", "(JJ)V");
  jlong sec= stamp.secPastEpoch;
  jlong nsec= stamp.nsec;
  return env->NewObject(clazz, init, sec, nsec);
}

jobjectArray newLabels(JNIEnv* env, short nbstrs, char strs[MAX_ENUM_STATES][MAX_ENUM_STRING_SIZE]) {
  jclass clazz= env->FindClass("java/lang/String");

  jobjectArray res= (jobjectArray)env->NewObjectArray(nbstrs, clazz, NULL);

  for (int t=0; t<nbstrs; ++t) {
    env->SetObjectArrayElement(res, t, env->NewStringUTF(strs[t]));
  }

  return res;
}


jstring newUnits(JNIEnv* env, char units[MAX_UNITS_SIZE]) {
  return env->NewStringUTF(units);
}




JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1create(JNIEnv* env, jclass, jint type, jint length) {
  return(jint) malloc(dbr_size_n(type,length));
}

JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1destroy(JNIEnv* env, jclass, jint dbrid) {
  free((void*)dbrid);
}


JNIEXPORT void JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1setValue(JNIEnv *env, jclass, jint dbrid, jint type, jint length, jobject value) {
  void* pvalue= dbr_value_ptr(dbrid,type);

  if (dbr_type_is_CHAR(type)) {
    env->GetByteArrayRegion((jbyteArray)value, 0,length, (jbyte*)pvalue);
  } else if (dbr_type_is_SHORT(type)) {
    env->GetShortArrayRegion((jshortArray)value, 0,length, (jshort*)pvalue);
  } else if (dbr_type_is_LONG(type)) {
    env->GetIntArrayRegion((jintArray)value, 0,length, (jint*)pvalue);
  } else if (dbr_type_is_FLOAT(type)) {
    env->GetFloatArrayRegion((jfloatArray)value, 0,length, (jfloat*)pvalue);
  } else if (dbr_type_is_DOUBLE(type)) {
    env->GetDoubleArrayRegion((jdoubleArray)value, 0,length, (jdouble*)pvalue);
  } else if (dbr_type_is_STRING(type)) {
    dbr_string_t* pstrs= (dbr_string_t*)pvalue;
    const char* str;
    for (int t=0; t<length; ++t) {
      str= env->GetStringUTFChars((jstring) env->GetObjectArrayElement((jobjectArray)value,t) , NULL);
      memcpy(pstrs[t],str,MAX_STRING_SIZE);
      pstrs[t][MAX_STRING_SIZE-1]=0;
    }
  }
}



JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getValue(JNIEnv *env, jclass, jint dbrid, jint type, jint length, jobject value) {
  void* pvalue= dbr_value_ptr(dbrid,type);

  if (value==NULL) {
    if (dbr_type_is_CHAR(type)) {
      value= env->NewByteArray(length);
    } else if (dbr_type_is_SHORT(type)) {
      value= env->NewShortArray(length);
    } else if (dbr_type_is_LONG(type)) {
      value= env->NewIntArray(length);
    } else if (dbr_type_is_FLOAT(type)) {
      value= env->NewFloatArray(length);
    } else if (dbr_type_is_DOUBLE(type)) {
      value= env->NewDoubleArray(length);
    } else if (dbr_type_is_STRING(type)) {
      value= env->NewObjectArray(length, env->FindClass("java/lang/String"), NULL);
    }
  }

  if (dbr_type_is_CHAR(type)) {
    env->SetByteArrayRegion((jbyteArray)value, 0,length, (jbyte*)pvalue);
  } else if (dbr_type_is_SHORT(type)) {
    env->SetShortArrayRegion((jshortArray)value, 0,length, (jshort*)pvalue);
  } else if (dbr_type_is_LONG(type)) {
    env->SetIntArrayRegion((jintArray)value, 0,length, (jint*)pvalue);
  } else if (dbr_type_is_FLOAT(type)) {
    env->SetFloatArrayRegion((jfloatArray)value, 0,length, (jfloat*)pvalue);
  } else if (dbr_type_is_DOUBLE(type)) {
    env->SetDoubleArrayRegion((jdoubleArray)value, 0,length, (jdouble*)pvalue);
  } else if (dbr_type_is_STRING(type)) {
    dbr_string_t* pstrs= (dbr_string_t*)pvalue;
    jstring str;
    for (int t=0; t<length; ++t) {
      str= env->NewStringUTF(pstrs[t]);
      env->SetObjectArrayElement((jobjectArray)value,t, str);
    }
  }

  return value;
}



JNIEXPORT jshort JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getStatus(JNIEnv* env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_STS_STRING:  return(jshort) ((dbr_sts_string*)dbrid)->status;
  case DBR_TIME_STRING: return(jshort) ((dbr_time_string*)dbrid)->status;

  case DBR_STS_ENUM:    return(jshort) ((dbr_sts_enum*)dbrid)->status;
  case DBR_TIME_ENUM:   return(jshort) ((dbr_time_enum*)dbrid)->status;
  case DBR_GR_ENUM:     return(jshort) ((dbr_gr_enum*)dbrid)->status;
  case DBR_CTRL_ENUM:   return(jshort) ((dbr_ctrl_enum*)dbrid)->status;

  case DBR_STS_CHAR:    return(jshort) ((dbr_sts_char*)dbrid)->status;
  case DBR_TIME_CHAR:   return(jshort) ((dbr_time_char*)dbrid)->status;
  case DBR_GR_CHAR:     return(jshort) ((dbr_gr_char*)dbrid)->status;
  case DBR_CTRL_CHAR:   return(jshort) ((dbr_ctrl_char*)dbrid)->status;

  case DBR_STS_SHORT:   return(jshort) ((dbr_sts_short*)dbrid)->status;
  case DBR_TIME_SHORT:  return(jshort) ((dbr_time_short*)dbrid)->status;
  case DBR_GR_SHORT:    return(jshort) ((dbr_gr_short*)dbrid)->status;
  case DBR_CTRL_SHORT:  return(jshort) ((dbr_ctrl_short*)dbrid)->status;

  case DBR_STS_LONG:    return(jshort) ((dbr_sts_long*)dbrid)->status;
  case DBR_TIME_LONG:   return(jshort) ((dbr_time_long*)dbrid)->status;
  case DBR_GR_LONG:     return(jshort) ((dbr_gr_long*)dbrid)->status;
  case DBR_CTRL_LONG:   return(jshort) ((dbr_ctrl_long*)dbrid)->status;

  case DBR_STS_FLOAT:   return(jshort) ((dbr_sts_float*)dbrid)->status;
  case DBR_TIME_FLOAT:  return(jshort) ((dbr_time_float*)dbrid)->status;
  case DBR_GR_FLOAT:    return(jshort) ((dbr_gr_float*)dbrid)->status;
  case DBR_CTRL_FLOAT:  return(jshort) ((dbr_ctrl_float*)dbrid)->status;

  case DBR_STS_DOUBLE:  return(jshort) ((dbr_sts_double*)dbrid)->status;
  case DBR_TIME_DOUBLE: return(jshort) ((dbr_time_double*)dbrid)->status;
  case DBR_GR_DOUBLE:   return(jshort) ((dbr_gr_double*)dbrid)->status;
  case DBR_CTRL_DOUBLE: return(jshort) ((dbr_ctrl_double*)dbrid)->status;
  }
  return(jshort)0;
}


JNIEXPORT jshort JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getSeverity(JNIEnv* env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_STS_STRING:  return(jshort) ((dbr_sts_string*)dbrid)->severity;
  case DBR_TIME_STRING: return(jshort) ((dbr_time_string*)dbrid)->severity;

  case DBR_STS_ENUM:    return(jshort) ((dbr_sts_enum*)dbrid)->severity;
  case DBR_TIME_ENUM:   return(jshort) ((dbr_time_enum*)dbrid)->severity;
  case DBR_GR_ENUM:     return(jshort) ((dbr_gr_enum*)dbrid)->severity;
  case DBR_CTRL_ENUM:   return(jshort) ((dbr_ctrl_enum*)dbrid)->severity;

  case DBR_STS_CHAR:    return(jshort) ((dbr_sts_char*)dbrid)->severity;
  case DBR_TIME_CHAR:   return(jshort) ((dbr_time_char*)dbrid)->severity;
  case DBR_GR_CHAR:     return(jshort) ((dbr_gr_char*)dbrid)->severity;
  case DBR_CTRL_CHAR:   return(jshort) ((dbr_ctrl_char*)dbrid)->severity;

  case DBR_STS_SHORT:   return(jshort) ((dbr_sts_short*)dbrid)->severity;
  case DBR_TIME_SHORT:  return(jshort) ((dbr_time_short*)dbrid)->severity;
  case DBR_GR_SHORT:    return(jshort) ((dbr_gr_short*)dbrid)->severity;
  case DBR_CTRL_SHORT:  return(jshort) ((dbr_ctrl_short*)dbrid)->severity;

  case DBR_STS_LONG:    return(jshort) ((dbr_sts_long*)dbrid)->severity;
  case DBR_TIME_LONG:   return(jshort) ((dbr_time_long*)dbrid)->severity;
  case DBR_GR_LONG:     return(jshort) ((dbr_gr_long*)dbrid)->severity;
  case DBR_CTRL_LONG:   return(jshort) ((dbr_ctrl_long*)dbrid)->severity;

  case DBR_STS_FLOAT:   return(jshort) ((dbr_sts_float*)dbrid)->severity;
  case DBR_TIME_FLOAT:  return(jshort) ((dbr_time_float*)dbrid)->severity;
  case DBR_GR_FLOAT:    return(jshort) ((dbr_gr_float*)dbrid)->severity;
  case DBR_CTRL_FLOAT:  return(jshort) ((dbr_ctrl_float*)dbrid)->severity;

  case DBR_STS_DOUBLE:  return(jshort) ((dbr_sts_double*)dbrid)->severity;
  case DBR_TIME_DOUBLE: return(jshort) ((dbr_time_double*)dbrid)->severity;
  case DBR_GR_DOUBLE:   return(jshort) ((dbr_gr_double*)dbrid)->severity;
  case DBR_CTRL_DOUBLE: return(jshort) ((dbr_ctrl_double*)dbrid)->severity;
  }
  return(jshort)0;
}




JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getUDL(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_GR_CHAR:     return newByte(env, ((dbr_gr_char*)dbrid)->upper_disp_limit);
  case DBR_CTRL_CHAR:   return newByte(env, ((dbr_ctrl_char*)dbrid)->upper_disp_limit);

  case DBR_GR_SHORT:    return newShort(env, ((dbr_gr_short*)dbrid)->upper_disp_limit);
  case DBR_CTRL_SHORT:  return newShort(env, ((dbr_ctrl_short*)dbrid)->upper_disp_limit);

  case DBR_GR_LONG:     return newInt(env, ((dbr_gr_long*)dbrid)->upper_disp_limit);
  case DBR_CTRL_LONG:   return newInt(env, ((dbr_ctrl_long*)dbrid)->upper_disp_limit);

  case DBR_GR_FLOAT:    return newFloat(env, ((dbr_gr_float*)dbrid)->upper_disp_limit);
  case DBR_CTRL_FLOAT:  return newFloat(env, ((dbr_ctrl_float*)dbrid)->upper_disp_limit);

  case DBR_GR_DOUBLE:   return newDouble(env, ((dbr_gr_double*)dbrid)->upper_disp_limit);
  case DBR_CTRL_DOUBLE: return newDouble(env, ((dbr_ctrl_double*)dbrid)->upper_disp_limit);
  }
  return NULL;
}


JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getLDL(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_GR_CHAR:     return newByte(env, ((dbr_gr_char*)dbrid)->lower_disp_limit);
  case DBR_CTRL_CHAR:   return newByte(env, ((dbr_ctrl_char*)dbrid)->lower_disp_limit);

  case DBR_GR_SHORT:    return newShort(env, ((dbr_gr_short*)dbrid)->lower_disp_limit);
  case DBR_CTRL_SHORT:  return newShort(env, ((dbr_ctrl_short*)dbrid)->lower_disp_limit);

  case DBR_GR_LONG:     return newInt(env, ((dbr_gr_long*)dbrid)->lower_disp_limit);
  case DBR_CTRL_LONG:   return newInt(env, ((dbr_ctrl_long*)dbrid)->lower_disp_limit);

  case DBR_GR_FLOAT:    return newFloat(env, ((dbr_gr_float*)dbrid)->lower_disp_limit);
  case DBR_CTRL_FLOAT:  return newFloat(env, ((dbr_ctrl_float*)dbrid)->lower_disp_limit);

  case DBR_GR_DOUBLE:   return newDouble(env, ((dbr_gr_double*)dbrid)->lower_disp_limit);
  case DBR_CTRL_DOUBLE: return newDouble(env, ((dbr_ctrl_double*)dbrid)->lower_disp_limit);
  }
  return NULL;
}

JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getUAL(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_GR_CHAR:     return newByte(env, ((dbr_gr_char*)dbrid)->upper_alarm_limit);
  case DBR_CTRL_CHAR:   return newByte(env, ((dbr_ctrl_char*)dbrid)->upper_alarm_limit);

  case DBR_GR_SHORT:    return newShort(env, ((dbr_gr_short*)dbrid)->upper_alarm_limit);
  case DBR_CTRL_SHORT:  return newShort(env, ((dbr_ctrl_short*)dbrid)->upper_alarm_limit);

  case DBR_GR_LONG:     return newInt(env, ((dbr_gr_long*)dbrid)->upper_alarm_limit);
  case DBR_CTRL_LONG:   return newInt(env, ((dbr_ctrl_long*)dbrid)->upper_alarm_limit);

  case DBR_GR_FLOAT:    return newFloat(env, ((dbr_gr_float*)dbrid)->upper_alarm_limit);
  case DBR_CTRL_FLOAT:  return newFloat(env, ((dbr_ctrl_float*)dbrid)->upper_alarm_limit);

  case DBR_GR_DOUBLE:   return newDouble(env, ((dbr_gr_double*)dbrid)->upper_alarm_limit);
  case DBR_CTRL_DOUBLE: return newDouble(env, ((dbr_ctrl_double*)dbrid)->upper_alarm_limit);
  }
  return NULL;
}


JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getUWL(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_GR_CHAR:     return newByte(env, ((dbr_gr_char*)dbrid)->upper_warning_limit);
  case DBR_CTRL_CHAR:   return newByte(env, ((dbr_ctrl_char*)dbrid)->upper_warning_limit);

  case DBR_GR_SHORT:    return newShort(env, ((dbr_gr_short*)dbrid)->upper_warning_limit);
  case DBR_CTRL_SHORT:  return newShort(env, ((dbr_ctrl_short*)dbrid)->upper_warning_limit);

  case DBR_GR_LONG:     return newInt(env, ((dbr_gr_long*)dbrid)->upper_warning_limit);
  case DBR_CTRL_LONG:   return newInt(env, ((dbr_ctrl_long*)dbrid)->upper_warning_limit);

  case DBR_GR_FLOAT:    return newFloat(env, ((dbr_gr_float*)dbrid)->upper_warning_limit);
  case DBR_CTRL_FLOAT:  return newFloat(env, ((dbr_ctrl_float*)dbrid)->upper_warning_limit);

  case DBR_GR_DOUBLE:   return newDouble(env, ((dbr_gr_double*)dbrid)->upper_warning_limit);
  case DBR_CTRL_DOUBLE: return newDouble(env, ((dbr_ctrl_double*)dbrid)->upper_warning_limit);
  }
  return NULL;
}


JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getLWL(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_GR_CHAR:     return newByte(env, ((dbr_gr_char*)dbrid)->lower_warning_limit);
  case DBR_CTRL_CHAR:   return newByte(env, ((dbr_ctrl_char*)dbrid)->lower_warning_limit);

  case DBR_GR_SHORT:    return newShort(env, ((dbr_gr_short*)dbrid)->lower_warning_limit);
  case DBR_CTRL_SHORT:  return newShort(env, ((dbr_ctrl_short*)dbrid)->lower_warning_limit);

  case DBR_GR_LONG:     return newInt(env, ((dbr_gr_long*)dbrid)->lower_warning_limit);
  case DBR_CTRL_LONG:   return newInt(env, ((dbr_ctrl_long*)dbrid)->lower_warning_limit);

  case DBR_GR_FLOAT:    return newFloat(env, ((dbr_gr_float*)dbrid)->lower_warning_limit);
  case DBR_CTRL_FLOAT:  return newFloat(env, ((dbr_ctrl_float*)dbrid)->lower_warning_limit);

  case DBR_GR_DOUBLE:   return newDouble(env, ((dbr_gr_double*)dbrid)->lower_warning_limit);
  case DBR_CTRL_DOUBLE: return newDouble(env, ((dbr_ctrl_double*)dbrid)->lower_warning_limit);
  }
  return NULL;
}


JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getLAL(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_GR_CHAR:     return newByte(env, ((dbr_gr_char*)dbrid)->lower_alarm_limit);
  case DBR_CTRL_CHAR:   return newByte(env, ((dbr_ctrl_char*)dbrid)->lower_alarm_limit);

  case DBR_GR_SHORT:    return newShort(env, ((dbr_gr_short*)dbrid)->lower_alarm_limit);
  case DBR_CTRL_SHORT:  return newShort(env, ((dbr_ctrl_short*)dbrid)->lower_alarm_limit);

  case DBR_GR_LONG:     return newInt(env, ((dbr_gr_long*)dbrid)->lower_alarm_limit);
  case DBR_CTRL_LONG:   return newInt(env, ((dbr_ctrl_long*)dbrid)->lower_alarm_limit);

  case DBR_GR_FLOAT:    return newFloat(env, ((dbr_gr_float*)dbrid)->lower_alarm_limit);
  case DBR_CTRL_FLOAT:  return newFloat(env, ((dbr_ctrl_float*)dbrid)->lower_alarm_limit);

  case DBR_GR_DOUBLE:   return newDouble(env, ((dbr_gr_double*)dbrid)->lower_alarm_limit);
  case DBR_CTRL_DOUBLE: return newDouble(env, ((dbr_ctrl_double*)dbrid)->lower_alarm_limit);
  }
  return NULL;
}


JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getUCL(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_CTRL_CHAR:   return newByte(env, ((dbr_ctrl_char*)dbrid)->upper_ctrl_limit);

  case DBR_CTRL_SHORT:  return newShort(env, ((dbr_ctrl_short*)dbrid)->upper_ctrl_limit);

  case DBR_CTRL_LONG:   return newInt(env, ((dbr_ctrl_long*)dbrid)->upper_ctrl_limit);

  case DBR_CTRL_FLOAT:  return newFloat(env, ((dbr_ctrl_float*)dbrid)->upper_ctrl_limit);

  case DBR_CTRL_DOUBLE: return newDouble(env, ((dbr_ctrl_double*)dbrid)->upper_ctrl_limit);
  }
  return NULL;
}


JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getLCL(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_CTRL_CHAR:   return newByte(env, ((dbr_ctrl_char*)dbrid)->lower_ctrl_limit);

  case DBR_CTRL_SHORT:  return newShort(env, ((dbr_ctrl_short*)dbrid)->lower_ctrl_limit);

  case DBR_CTRL_LONG:   return newInt(env, ((dbr_ctrl_long*)dbrid)->lower_ctrl_limit);

  case DBR_CTRL_FLOAT:  return newFloat(env, ((dbr_ctrl_float*)dbrid)->lower_ctrl_limit);

  case DBR_CTRL_DOUBLE: return newDouble(env, ((dbr_ctrl_double*)dbrid)->lower_ctrl_limit);
  }
  return NULL;
}

JNIEXPORT jshort JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getPrecision(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  case DBR_GR_FLOAT:    return(jshort) ((dbr_gr_float*)dbrid)->precision;
  case DBR_CTRL_FLOAT:  return(jshort) ((dbr_ctrl_float*)dbrid)->precision;
  case DBR_GR_DOUBLE:   return(jshort) ((dbr_gr_double*)dbrid)->precision;
  case DBR_CTRL_DOUBLE: return(jshort) ((dbr_ctrl_double*)dbrid)->precision;
  }
  return(jshort)0;
}


JNIEXPORT jobject JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getTimeStamp(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_TIME_STRING: return newTimeStamp(env, ((dbr_time_string*)dbrid)->stamp);

  case DBR_TIME_ENUM:   return newTimeStamp(env, ((dbr_time_enum*)dbrid)->stamp);

  case DBR_TIME_CHAR:   return newTimeStamp(env, ((dbr_time_char*)dbrid)->stamp);

  case DBR_TIME_SHORT:  return newTimeStamp(env, ((dbr_time_short*)dbrid)->stamp);

  case DBR_TIME_LONG:   return newTimeStamp(env, ((dbr_time_long*)dbrid)->stamp);

  case DBR_TIME_FLOAT:  return newTimeStamp(env, ((dbr_time_float*)dbrid)->stamp);

  case DBR_TIME_DOUBLE: return newTimeStamp(env, ((dbr_time_double*)dbrid)->stamp);

  }
  return NULL;
}


JNIEXPORT jobjectArray JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getLabels(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  case DBR_GR_ENUM:   return newLabels(env, ((dbr_gr_enum*)dbrid)->no_str, ((dbr_gr_enum*)dbrid)->strs);
  case DBR_CTRL_ENUM: return newLabels(env, ((dbr_ctrl_enum*)dbrid)->no_str, ((dbr_ctrl_enum*)dbrid)->strs);
  }
  return NULL;
}


JNIEXPORT jstring JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getUnits(JNIEnv *env, jclass, jint dbrid, jint type) {
  switch (type) {
  
  case DBR_GR_CHAR:   return newUnits(env, ((dbr_gr_char*)dbrid)->units);
  case DBR_CTRL_CHAR: return newUnits(env, ((dbr_ctrl_char*)dbrid)->units);

  case DBR_GR_SHORT:   return newUnits(env, ((dbr_gr_short*)dbrid)->units);
  case DBR_CTRL_SHORT: return newUnits(env, ((dbr_ctrl_short*)dbrid)->units);

  case DBR_GR_LONG:    return newUnits(env, ((dbr_gr_long*)dbrid)->units);
  case DBR_CTRL_LONG:  return newUnits(env, ((dbr_ctrl_long*)dbrid)->units);

  case DBR_GR_FLOAT:   return newUnits(env, ((dbr_gr_float*)dbrid)->units);
  case DBR_CTRL_FLOAT: return newUnits(env, ((dbr_ctrl_float*)dbrid)->units);

  case DBR_GR_DOUBLE:   return newUnits(env, ((dbr_gr_double*)dbrid)->units);
  case DBR_CTRL_DOUBLE: return newUnits(env, ((dbr_ctrl_double*)dbrid)->units);

  }
  return NULL;
}

JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getAckT(JNIEnv *env, jclass, jint dbrid, jint type) {
  if (type==DBR_STSACK_STRING) return((dbr_stsack_string*)dbrid)->ackt;
  return 0;
}

JNIEXPORT jint JNICALL Java_gov_aps_epics_jca_jni_JNI__1dbr_1getAckS(JNIEnv *env, jclass, jint dbrid, jint type) {
  if (type==DBR_STSACK_STRING) return((dbr_stsack_string*)dbrid)->acks;
  return 0;
}

